package mt.deepnight.slb;

import mt.MLib;

typedef AnimInstance = {
	var group			: String;
	var frames			: Array<Int>;
	var animCursor		: Int;
	var curFrameCpt		: Float;
	var plays			: Int;
	var paused			: Bool;
	var isStateAnim		: Bool;
	var killAfterPlay	: Bool;
	var onEnd			: Null< Void->Void >;
	var onLoop			: Null< Void->Void >;
}

class AnimManager {
	public static var UPDATE_LIST : List<AnimManager> = new List();

	var speed		: Float;
	var animStack	: Array<AnimInstance>;
	var spr			: BSprite;
	var stateAnims	: Array<{group:String, priority:Int, condition:Void->Bool}>;
	var suspendTime	: Int;
	var destroyed	: Bool;

	var isPlaying	: Bool;

	public function new(spr:BSprite) {
		this.spr = spr;
		animStack = [];
		stateAnims = [];
		suspendTime = 0;
		speed = 1;
		destroyed = false;
		isPlaying = false;
	}

	inline function getCurrentAnim() {
		return animStack[0];
	}

	inline function getLastAnim() {
		return animStack[animStack.length-1];
	}

	function startUpdates() {
		UPDATE_LIST.remove(this);
		UPDATE_LIST.add(this);
	}

	function stopUpdates() {
		UPDATE_LIST.remove(this);
	}

	public function destroy() {
		destroyed = true;
		stopWithoutStateAnims();
		stopUpdates();
		stateAnims = null;
		animStack = null;
		spr = null;
	}

	public static function getUpdateListContent() {
		return
			UPDATE_LIST.length+" element(s): " +
			UPDATE_LIST.map(function(a) return a.spr.groupName).join(", ");
	}

	inline function hasAnim() {
		return animStack.length>0;
	}

	public inline function isPlayingAnim(?group:String) {
		return hasAnim() && !isSuspended() && (group==null || getCurrentAnim().group==group);
		//return hasAnim() && (group==null || getCurrentAnim().group==group);
	}

	public inline function getAnimCursor() {
		return hasAnim() ? getCurrentAnim().animCursor : 0;
	}

	public function chain(id:String) {
		play(id, true);
		return this;
	}

	public inline function chainAndLoop(k:String) {
		return chain(k).loop();
	}

	public inline function playAndLoop(k:String) {
		return play(k).loop();
	}

	public function play(group:String, ?plays=1, ?queueAnim=false) {
		var aframes = spr.lib.getGroup(group).anim;
		if( aframes==null && aframes.length>0 )
			return this;

		if( !queueAnim && hasAnim() )
			stopWithoutStateAnims();

		var a : AnimInstance = {
			group		: group,
			paused		: false,
			curFrameCpt	: 0,
			animCursor	: 0,
			plays		: plays,
			frames		: aframes,
			isStateAnim	: false,
			killAfterPlay: false,
			onEnd		: null,
			onLoop		: null,
		}
		isPlaying = true;
		animStack.push(a);
		startUpdates();

		if( !queueAnim )
			initCurrentAnim();

		return this;
	}

	public function loop() {
		if( hasAnim() )
			getLastAnim().plays = 999999;
		return this;
	}

	public function killAfterPlay() {
		if( hasAnim() )
			getCurrentAnim().killAfterPlay = true;
		return this;
	}

	public function stopOnEnd() {
		if( hasAnim() )
			getLastAnim().onEnd = function() stopWithoutStateAnims();
		return this;
	}

	public function onEnd(cb:Void->Void) {
		if( hasAnim() )
			getLastAnim().onEnd = cb;
		return this;
	}

	static var UNSYNC : Map<String,Int> = new Map();
	public function unsync() {
		if( !hasAnim() )
			return this;

		var a = getCurrentAnim();
		if( !UNSYNC.exists(a.group) )
			UNSYNC.set(a.group, 1);
		else
			UNSYNC.set(a.group, UNSYNC.get(a.group)+1);

		var offset = MLib.ceil(a.frames.length/3);
		a.animCursor = ( offset * UNSYNC.get(a.group) + Std.random(100) ) % a.frames.length;
		return this;
	}

	public function onEachLoop(cb:Void->Void) {
		if( hasAnim() )
			getLastAnim().onLoop = cb;
		return this;
	}

	public function pause() {
		if( hasAnim() )
			getCurrentAnim().paused = true;
	}

	public function resume() {
		if( hasAnim() )
			getCurrentAnim().paused = false;
	}

	public function stop() {
		isPlaying = false;
		animStack.splice(0, animStack.length);
		applyStateAnims();
	}

	public function stopWith(group:String, ?frame=0) {
		stopWithoutStateAnims();
		spr.set(group, frame);
	}

	public function stopWithoutStateAnims() {
		isPlaying = false;
		animStack.splice(0, animStack.length);
	}

	public function suspendFor(durationFrame:Int, ?group:String, ?frame=0) {
		suspendTime = durationFrame + 1;
		if( group!=null )
			spr.set(group, frame);
	}

	inline function isSuspended() {
		return suspendTime>0;
	}


	public function setSpeed(s:Float) {
		speed = s;
	}

	function nextAnim() {
		if( getCurrentAnim().killAfterPlay ) {
			spr.destroy();
			return;
		}

		animStack.shift();
		if( animStack.length==0 ) {
			// End
			stop();
		}
		else {
			// Init next
			initCurrentAnim();
		}
	}

	function initCurrentAnim() {
		var a = getCurrentAnim();
		spr.set(a.group, a.frames[0]);
	}



	public function registerStateAnim(group:String, priority:Int, ?condition:Void->Bool) {
		if( condition==null )
			condition = function() return true;

		removeStateAnim(group, priority);
		stateAnims.push({
			group		: group,
			priority	: priority,
			condition	: condition,
		});
		stateAnims.sort( function(a,b) return -Reflect.compare(a.priority, b.priority) );
		applyStateAnims();
	}

	public function removeStateAnim(group:String, priority:Int) {
		var i = 0;
		while( i<stateAnims.length )
			if( stateAnims[i].group==group && stateAnims[i].priority==priority )
				stateAnims.splice(i,1);
			else
				i++;
	}

	public function removeAllStateAnims() {
		stateAnims.splice(0, stateAnims.length);
	}

	function applyStateAnims() {
		if( isSuspended() )
			return;

		if( hasAnim() && !getCurrentAnim().isStateAnim )
			return;

		for(sa in stateAnims)
			if( sa.condition() ) {
				if( hasAnim() && getCurrentAnim().group==sa.group )
					break;

				play(sa.group);
				if( hasAnim() )
					getLastAnim().isStateAnim = true;
				break;
			}
	}


	public function toString() {
		return
			"AnimManager("+spr+")" +
			(hasAnim() ? "Playing(stack="+animStack.length+")" : "NoAnim");
	}


	public function update() {
		// Suspended
		if( isSuspended() ) {
			suspendTime--;
			if( isSuspended() )
				return;
			else {
				suspendTime = 0;
				if( hasAnim() ) {
					var a = getCurrentAnim();
					spr.set( a.group, a.frames[a.animCursor] ); // back to previous state
				}
			}
		}

		// State anims
		applyStateAnims();

		// Playback
		var a = getCurrentAnim();
		if( a!=null && !a.paused ) {
			a.curFrameCpt += speed;

			while( a.curFrameCpt>1 ) {
				a.curFrameCpt--;
				a.animCursor++;
				if( a.animCursor>=a.frames.length ) {
					// Last frame reached
					a.animCursor = 0;
					a.plays--;
					if(a.plays<=0) {
						// Playback end
						if( a.onEnd!=null ) {
							var cb = a.onEnd;
							a.onEnd = null;
							cb();
						}
						if( hasAnim() )
							nextAnim();
					}
					else {
						// Loop
						if( a.onLoop!=null )
							a.onLoop();
						spr.setFrame( a.frames[a.animCursor] );
					}
				}
				else {
					// Normal frame
					spr.setFrame( a.frames[a.animCursor] );
				}
			}
		}

		// Nothing to do
		if( !destroyed && !hasAnim() && !isSuspended() )
			stopUpdates();
	}
}