package mt.deepnight.slb;

import mt.MLib;

#if !macro
import flash.display.Bitmap;
import flash.display.BitmapData;
import flash.geom.Point;
import flash.geom.Rectangle;
import haxe.xml.Fast;
#end

import haxe.macro.Expr;
import haxe.macro.Context;


#if !macro

//This should become a class for fast cpp rendering (typedef is slow)
typedef FrameData = {
	x			: Int,
	y			: Int,
	wid			: Int,
	hei			: Int,
	realFrame	: {x:Int, y:Int, realWid:Int, realHei:Int},
	rect		: Rectangle,
}

typedef LibGroup = {
	id		: String,
	maxWid	: Int,
	maxHei	: Int,
	frames	: Array<FrameData>,
	anim	: Array<Int>,
};
#end

enum SLBError {
	UnknownGroup(g:String);
	NoGroupSelected;
	GroupAlreadyExists(g:String);
	NoFrameInGroup(f:Int, g:String);
	WidthMismatch(f:Int, g:String);
	HeightMismatch(f:Int, g:String);
	InvalidFrameDuration(s:String);
	EndFrameLower(s:String);
	InvalidFrames(s:String);
	NoCurrentGroup;
	AnimFrameExceeds(id:String, anim:String, frame:Int);
	AssetImportFailed(e:Dynamic);
	NotSameSLBFromBatch;
}


/** LIBRARY ********************************************************************************************/

class BLib {
	#if !macro
	public var source			: BitmapData;
	var groups					: Map<String, LibGroup>;
	var currentGroup			: Null<LibGroup>;
	var frameRandDraw			: Map<String, Array<Int>>;
	public var defaultCenterX(default, null)	: Float;
	public var defaultCenterY(default, null)	: Float;
	var gridX					: Int;
	var gridY					: Int;

	#if h3d
	public var tile				: h2d.Tile;
	#end

	public function new(bd:BitmapData) {
		source = bd;
		groups = new Map();
		frameRandDraw = new Map();
		defaultCenterX = 0;
		defaultCenterY = 0;
		gridX = gridY = 16;

		#if h3d
		genTex();
		#end
	}

	#if h3d
	function genTex() {
		if( h3d.Engine.getCurrent() != null )
			tile = h2d.Tile.fromBitmap( hxd.BitmapData.fromNative(source) );
	}

	/**
	 * Get a Batch from a BSprite.
	 * @param	b:SpriteBatch
	 * @param	idSprite:String
	 * @param	frame:Int
	 * @return
	 */
	public function getBatchElement(b:h2d.SpriteBatch, idSprite:String, ?frame:Int = 0, ?d:{x:Float, y:Float} = null, ?p:{x:Int, y:Int} = null):h2d.SpriteBatch.BatchElement {
		if (b.tile.getTexture() != tile.getTexture())
			throw SLBError.NotSameSLBFromBatch;

		var e = b.alloc(tile=tile.clone());
		var fd = this.getFrameData(idSprite, frame);

		e.t.setPos(fd.x, fd.y);
		e.t.setSize(fd.wid, fd.hei);
		e.t = e.t.center(fd.realFrame.x, fd.realFrame.y);

		return e;
	}
	#end

	public function destroy() {
		source.dispose();
		source = null;
	}


	public inline function get(k:String, ?frame=0) : BSprite {
		return new BSprite(this, k, frame);
	}

	#if h3d
	public inline function getTile(g:String, ?frame=0, px:Float=0.0,py:Float=0.0) : h2d.Tile {
		var fd = getFrameData(g, frame);
		var t = tile.clone();
		t.setPos(fd.x, fd.y);
		t.setSize(fd.wid, fd.hei);
		t.dx = -Std.int(fd.realFrame.realWid*px + fd.realFrame.x);
		t.dy = -Std.int(fd.realFrame.realHei*py + fd.realFrame.y);
		return t;
	}
	#end

	public function getAndPlay(k:String, ?plays=99999, ?killAfterPlay=false) : BSprite {
		var s = new BSprite(this);
		s.a.play(k, plays);
		if( killAfterPlay )
			s.a.killAfterPlay();
		return s;
	}


	public inline function setDefaultCenter(cx,cy) {
		defaultCenterX = cx;
		defaultCenterY = cy;
	}

	public function setSliceGrid(w,h) {
		gridX = w;
		gridY = h;
	}

	public inline function getGroup(?k:String) {
		return
			if(k==null ) {
				if( currentGroup==null )
					throw SLBError.NoGroupSelected;
				else
					currentGroup;
			}
			else
				if(groups.exists(k))
					groups.get(k);
				else
					throw SLBError.UnknownGroup(k);
	}

	public inline function getGroups() {
		return groups;
	}

	public inline function getAnim(k) {
		return getGroup(k).anim;
	}

	public inline function getAnimDuration(k) {
		return getAnim(k).length;
	}

	function createGroup(k:String) {
		if( groups.exists(k) )
			throw SLBError.GroupAlreadyExists(k);
		groups.set(k, {
			id		: k,
			maxWid	: 0,
			maxHei	: 0,
			frames	: new Array(),
			anim	: new Array(),
		});
		return setCurrentGroup(k);
	}

	inline function setCurrentGroup(k:String) {
		currentGroup = getGroup(k);
		return getGroup();
	}


	public inline function getRectangle(k:String, ?frame=0) {
		var g = getGroup(k);
		var fr = g.frames[frame];
		if ( fr == null) throw SLBError.NoFrameInGroup(frame, k);
		return new flash.geom.Rectangle(fr.x, fr.y, fr.wid, fr.hei);
	}

	public inline function getFrameData(k:String, ?frame=0) {
		var fr = getGroup(k).frames[frame];
		if ( fr == null) throw SLBError.NoFrameInGroup(frame, k);
		return fr;
	}

	public inline function exists(k:String,?frame=0) {
		return groups.exists(k) && groups.get(k).frames.length>frame;
	}

	public inline function getRandomFrame(k:String, ?randFunc:Int->Int) {
		if(randFunc==null)
			randFunc = Std.random;

		return
			if(frameRandDraw.exists(k)) {
				var a = frameRandDraw.get(k);
				a[ randFunc(a.length) ];
			}
			else
				randFunc(countFrames(k));
	}

	public inline function countFrames(k:String) {
		return getGroup(k).frames.length;
	}

	public inline function getRandom(k:String, ?randFunc:Int->Int) : BSprite {
		return get(k, getRandomFrame(k, randFunc));
	}

	public inline function getMovieClip(k:String, ?frame=0, ?centerX, ?centerY) : flash.display.MovieClip {
		var mc = new flash.display.MovieClip();
		drawIntoGraphics(mc.graphics, k, frame, centerX, centerY);
		return mc;
	}

	public inline function drawIntoGraphics(g:flash.display.Graphics, k:String, ?frame=0, ?centerX, ?centerY) {
		if(centerX==null)	centerX = defaultCenterX;
		if(centerY==null)	centerY = defaultCenterY;

		var fdata = getFrameData(k,frame);
		var m = new flash.geom.Matrix();
		m.translate( -fdata.realFrame.x-Std.int(fdata.realFrame.realWid*centerX), -fdata.realFrame.y-Std.int(fdata.realFrame.realHei*centerY) );
		//m.translate(Std.int(-rect.x - centerX*rect.width), Std.int(-rect.y - centerY*rect.height));
		g.beginBitmapFill(source, m, false, false);
		g.drawRect(Std.int(-centerX*fdata.rect.width), Std.int(-centerY*fdata.rect.height), fdata.rect.width, fdata.rect.height);
		g.endFill();
	}

	public inline function drawIntoBitmap(bd:flash.display.BitmapData, x:Float,y:Float, k:String, ?frame=0, ?centerX, ?centerY) {
		if(centerX==null)	centerX = defaultCenterX;
		if(centerY==null)	centerY = defaultCenterY;

		var fdata = getFrameData(k,frame);
		bd.copyPixels(
			source, fdata.rect,
			new flash.geom.Point(x-fdata.realFrame.x-Std.int(fdata.realFrame.realWid*centerX), y-fdata.realFrame.y-Std.int(fdata.realFrame.realHei*centerY)),
			true
		);
	}

	public inline function drawIntoBitmapRandom(bd:flash.display.BitmapData, x:Float,y:Float, k:String, ?randFunc:Int->Int, ?centerX, ?centerY) {
		drawIntoBitmap(bd, x,y,k, getRandomFrame(k, randFunc), centerX, centerY);
	}

	public function getBitmapData(k:String, ?frame=0, ?padding=0) {
		var fdata = getFrameData(k, frame);
		var bd = new BitmapData(Std.int(fdata.rect.width+padding*2), Std.int(fdata.rect.height+padding*2), true, 0x0);
		drawIntoBitmap(bd, padding,padding, k,frame, 0,0);
		return bd;
	}



	public function sliceCustom(groupName:String, frame:Int, x:Int, y:Int, wid:Int, hei:Int, ?realFrame:{x:Int, y:Int, realWid:Int, realHei:Int}) {
		var g = if( exists(groupName) ) getGroup(groupName) else createGroup(groupName);
		g.maxWid = MLib.max( g.maxWid, wid );
		g.maxHei = MLib.max( g.maxHei, hei );

		//if( wid!=g.wid )
			//throw SLBError.WidthMismatch(frame, groupName);
			//
		//if( hei!=g.hei )
			//throw SLBError.HeightMismatch(frame, groupName);
		if( realFrame==null )
			realFrame = {x:0, y:0, realWid:wid, realHei:hei}

		g.frames[frame] = { x:x, y:y, wid:wid, hei:hei, realFrame:realFrame, rect:new Rectangle(x,y,wid,hei) };
	}

	public function slice(groupName:String, x:Int, y:Int, wid:Int, hei:Int, ?repeatX=1, ?repeatY=1) {
		var g = createGroup(groupName);
		setCurrentGroup(groupName);
		g.maxWid = MLib.max( g.maxWid, wid );
		g.maxHei = MLib.max( g.maxHei, hei );
		for(iy in 0...repeatY)
			for(ix in 0...repeatX)
				g.frames.push({ x : x+ix*wid, y : y+iy*hei, wid:wid, hei:hei, realFrame:{x:0,y:0,realWid:wid,realHei:hei}, rect:new Rectangle(x+ix*wid,y+iy*hei,wid,hei) });
	}

	public function sliceGrid(groupName:String, gx:Int, gy:Int, ?repeatX=1, ?repeatY=1) {
		var g = createGroup(groupName);
		setCurrentGroup(groupName);
		g.maxWid = MLib.max( g.maxWid, gridX );
		g.maxHei = MLib.max( g.maxHei, gridY );
		for(iy in 0...repeatY)
			for(ix in 0...repeatX)
				g.frames.push({ x : gridX*(gx+ix), y : gridY*(gy+iy), wid:gridX, hei:gridY, realFrame:{x:0,y:0,realWid:gridX,realHei:gridY}, rect:new Rectangle(gridX*(gx+ix), gridY*(gy+iy), gridX, gridY) });
	}

	//public function sliceMore(x:Int, y:Int, ?repeatX=1, ?repeatY=1) {
		//var g = getGroup();
		//for(iy in 0...repeatY)
			//for(ix in 0...repeatX)
				//g.frames.push({ x : x+ix*g.wid, y : y+iy*g.hei, wid:g.wid, hei:g.hei });
	//}

	//public function applyPermanentFilter(k:String, filter:flash.filters.BitmapFilter) {
		//for( frame in 0...countFrames(k) ) {
			//var fdata = getFrameData(k, frame);
			//source.applyFilter(source, fdata.rect, fdata.rect.topLeft, filter);
		//}
	//}


	public function multiplayAllAnimDurations(factor:Int) {
		for(g in getGroups())
			if( g.anim.length>0 ) {
				var old = g.anim.copy();
				g.anim = [];
				for(f in old)
					for(i in 0...factor)
						g.anim.push(f);
			}
	}


	//public function offsetOneBasedFrames() {
		//for(g in groups) {
			//for(i in 1...g.frames.length)
				//g.frames[i-1] = g.frames[i];
			//g.frames.pop();
				//
			//for(i in 0...g.anim.length)
				//g.anim[i] = g.anim[i]-1;
		//}
	//}



	public function parseAttachPoints(colors:Array<Int>, groupName:String, dotSpriteSuffix:String) { // TODO classes plutôt qu'objets anonymes
		var attachPoints = new Map();
		var dots = new BSprite(this);

		// Usage: attachPoints.get(animId).get(frame).get(color).x/y

		//for( g in getGroups() ) {
			//var animId = g.id;
			//if( groupName.indexOf(dotSpriteSuffix)>0 )
				//continue;

			if( !exists(groupName + dotSpriteSuffix) )
				throw "ERROR: Missing dots for anim "+groupName;

			attachPoints.set(groupName, new Map());

			for( af in 0...getAnim(groupName).length ) {
				var sprFrame = getAnim(groupName)[af];
				var dotFrame = getAnim(groupName+dotSpriteSuffix)[af];
				attachPoints.get(groupName).set(sprFrame, new Map());

				// Apply frame
				dots.set(groupName+dotSpriteSuffix, dotFrame);
				var dotsFData = getFrameData(dots.groupName, dots.frame);

				// Parse colors positions
				for(col in colors) {
					var bounds = dots.getBitmapData().getColorBoundsRect(0xffffffff, Color.addAlphaF(col));
					var x = -dotsFData.realFrame.realWid*0.5 + (bounds.x - dotsFData.realFrame.x);
					var y = -dotsFData.realFrame.realHei + (bounds.y - dotsFData.realFrame.y);
					if( bounds.x!=0 || bounds.y==0 )
						attachPoints.get(groupName).get(sprFrame).set(col, { x:x, y:y });
				}
			}
		//}

		dots.destroy();

		//#if debug
		//for( k in ATTACH_POINTS.keys() )
			//for( f in ATTACH_POINTS.get(k).keys() )
				//for( col in ATTACH_POINTS.get(k).get(f).keys() )
					//trace( k+" frame="+f+" color="+Color.intToHex(col)+" pt="+ATTACH_POINTS.get(k).get(f).get(col) );
		//#end
	}



	public function toString() {
		var l = [];
		for( k in getGroups().keys() ) {
			var g = getGroup(k);
			l.push(k+" ("+g.maxWid+"x"+g.maxHei+") : "+g.frames.length+" frame(s), "+(g.anim.length==0 ? "noAnim" : "animated("+g.anim.length+"f)" ));
		}
		l.sort(function(a,b) return Reflect.compare(a,b));
		return l.join("\n");
	}

	#end

	public static function parseAnimDefinition(animDef:String,?timin=1) {
		animDef = StringTools.replace(animDef, ")", "(");
		var frames : Array<Int> = new Array();
		var parts = animDef.split(",");
		for (p in parts) {
			var curTiming = timin;
			if( p.indexOf("(")>0 ) {
				var t = Std.parseInt( p.split("(")[1] );
				if( Math.isNaN(t) )
					throw SLBError.InvalidFrameDuration(p);
				curTiming = t;
				p = p.substr( 0, p.indexOf("(") );
			}
			if( p.indexOf("-")<0 ) {
				var f = Std.parseInt(p);
				for(i in 0...curTiming)
					frames.push(f);
				continue;
			}
			if( p.indexOf("-")>0 ) {
				var from = Std.parseInt(p.split("-")[0]);
				var to = Std.parseInt(p.split("-")[1])+1;
				if( to<from )
					throw SLBError.EndFrameLower(p);
				while( from<to ) {
					for(i in 0...curTiming)
						frames.push(from);
					from++;
				}
				continue;
			}
			throw SLBError.InvalidFrames(p);
		}
		return frames;
	}


	public function __defineAnim(?group:String, anim:Array<Int>) {
		#if !macro
		if( currentGroup==null && group==null )
			throw SLBError.NoCurrentGroup;

		if( group!=null )
			setCurrentGroup(group);

		var a = [];
		for(f in anim) {
			if( f>=currentGroup.frames.length )
				throw SLBError.AnimFrameExceeds(currentGroup.id, "["+anim.join(",")+"] "+currentGroup.frames.length, f);
			a.push(f);
		}

		currentGroup.anim = a;
		#end
	}




	#if macro
	static function error( ?msg="", p : Position ) {
		haxe.macro.Context.error(msg, p);
	}
	#end


	/* MACRO : Animation declaration ********************************************************************************************
	 *
	 * SYNTAX 1: frame[(optional_duration)], frame[(optional_duration)], ...
	 * SYNTAX 2: begin-end[(optional_duration)], begin-end[(optional_duration)], ...
	 *
	 * EXAMPLE: defineAnim( "walk", "0-5, 6(2), 7(1)" );
	 */
	public macro function defineAnim(ethis:Expr, ?groupName:String, ?baseFrame:Int, ?animDefinition:Expr) : Expr {
		var p = ethis.pos;
		var def = animDefinition;

		function parseError(str:String) {
			error(str + "\nSYNTAX: frame[(duration)], frame[(duration)], begin-end([duration]), ...", def.pos);
		}

		// Parameters
		if( isNull(def) && groupName==null )
			error("missing anim declaration", p);

		if( isNull(def) && groupName!=null ) {
			def = {expr:EConst(CString(groupName)), pos:p};
			groupName = null;
		}

		if( baseFrame==null )
			baseFrame = 0;

		if( isNull(def) )
			parseError("animation definition is required");

		var str = switch(def.expr) {
			case EConst(c) :
				switch( c ) {
					case CString(s) : s;
					default :
						parseError("a constant string is required here");
						null;
				}
			default :
				parseError("a constant string is required here");
				null;
		}

		var frames : Array<Int> = try{ parseAnimDefinition(str); } catch(e:Dynamic) { parseError(e); null; }

		var eframes = Lambda.array(Lambda.map(frames, function(f) return { pos:p, expr:EConst( CInt(Std.string(f)) ) } ));
		var arrayExpr = { pos:p, expr:EArrayDecl(eframes) };

		if( groupName==null )
			return macro $ethis.__defineAnim($arrayExpr);
		else {
			var groupExpr = { pos:p, expr:EConst(CString(groupName)) };
			return macro $ethis.__defineAnim($groupExpr, $arrayExpr);
		}
	}


	#if macro
	static function isNull(e:Expr) {
		switch(e.expr) {
			case EConst(c) :
				switch( c ) {
					case CIdent(v) : return v=="null";
					default :
						return false;
				}
			default :
				return false;
		}
	}
	#end
}



